(* Lambda Terms *)

signature LTERM =
sig

  structure Type : TYPE

  datatype term =
      Var of int		   (* w or u, as de Bruijn index 'n, n >= 1 *)
    | Tensor of term * term	   (* M1 # M2 *)
    | LetTensor of term * term	   (* let u1 # u2 = M in N *)
    | Star			   (* * *)
    | LetStar of term * term	   (* let * = M in N *)
    | LLam of term		   (* llam w. M *)
    | LApp of term * term	   (* M ^ N *)
    | Pair of term * term	   (* <M, N> *)
    | Fst of term		   (* fst M *)
    | Snd of term		   (* snd M *)
    | Unit			   (* <> *)
    | Inl of term		   (* inl M *)
    | Inr of term		   (* inr M *)
    | Case of term * term * term   (* case M of inl w1 => N1 | inr w2 => N2 *)
    | ULam of term		   (* lam u. M *)
    | UApp of term * term	   (* M N *)
    | Bang of term		   (* !M *)
    | LetBang of term * term	   (* let !u = M in N *)
    | Fold of term		   (* fold M *)
    | Unfold of term		   (* unfold M *)
    | Fix of term		   (* fix M *)
    | LetName of Type.tp * term * term  (* let name u = M in N *)

  (* subst (M, N) = [M/w]N or [M/u]N, where w or u are represented *)
  (* by de Bruijn index '1.  M must be closed and N have exactly *)
  (* one free variable *)
  val subst : term * term -> term

end;  (* signature LTERM *)
