structure Type :> TYPE =
struct

  type var = int

  datatype tp =				(* A ::= *)
      Var of var			(* 'n, n >= 1 *)
    | Tensor of tp * tp			(* A1 # A2 *)
    | One				(* 1 *)
    | Lolli of tp * tp			(* A1 -o A2 *)
    | With of tp * tp			(* A1 & A2 *)
    | Top				(* T *)
    | Plus of tp * tp			(* A1 + A2 *)
    | Zero				(* 0 *)
    | Arrow of tp * tp			(* A1 -> A2 *)
    | Bang of tp			(* !A *)
    | Mu of tp				(* mu 'a. A *)

  (* val eq : tp * tp -> bool *)
  fun eq (Var (j), Var (k)) = (j = k)
    | eq (Tensor (A1, A2), Tensor (B1, B2)) = eq (A1, B1) andalso eq (A2, B2)
    | eq (One, One) = true
    | eq (Lolli (A1, A2), Lolli (B1, B2)) = eq (A1, B1) andalso eq (A2, B2)
    | eq (With (A1, A2), With (B1, B2)) = eq (A1, B1) andalso eq (A2, B2)
    | eq (Top, Top) = true
    | eq (Plus (A1, A2), Plus (B1, B2)) = eq (A1, B1) andalso eq (A2, B2)
    | eq (Zero, Zero) = true
    | eq (Arrow (A1, A2), Arrow (B1, B2)) = eq (A1, B1) andalso eq (A2, B2)
    | eq (Bang (A1), Bang (B1)) = eq (A1, B1)
    | eq (Mu (A1), Mu (B1)) = eq (A1, B1)
    | eq (_, _) = false

  fun subst (A, B) =
      let
	fun sb (i, B as Var(j)) = if i = j then A else B
	  | sb (i, Tensor (B1,B2)) = Tensor (sb (i, B1), sb (i, B2))
	  | sb (i, One) = One
	  | sb (i, Lolli (B1, B2)) = Lolli (sb (i, B1), sb (i, B2))
	  | sb (i, With (B1, B2)) = With (sb (i, B1), sb (i, B2))
	  | sb (i, Top) = Top
	  | sb (i, Plus (B1, B2)) = Plus (sb (i, B1), sb (i, B2))
	  | sb (i, Zero) = Zero
	  | sb (i, Arrow (B1, B2)) = Arrow (sb (i, B1), sb (i, B2))
	  | sb (i, Bang (B1)) = Bang (sb (i, B1))
	  | sb (i, Mu (B1)) = Mu (sb (i+1, B1))
      in
	sb (1, B)
      end

end;  (* structure Type *)
